import datetime as dt
import logging
import os
import warnings
from typing import Literal

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import requests
import rioxarray
import shapely
import xarray as xr
from rasterio import merge
from rasterio.io import MemoryFile
from requests.exceptions import HTTPError
from rioxarray.merge import merge_arrays
from tqdm import tqdm

from .. import NLMOD_DATADIR, cache
from ..dims.grid import get_extent
from ..dims.resample import structured_da_to_ds
from ..util import extent_to_polygon, get_ds_empty
from .webservices import arcrest, wcs

logger = logging.getLogger(__name__)


@cache.cache_netcdf(coords_2d=True)
def download_ahn(
    extent: list[float],
    identifier: str = "AHN4 maaiveldmodel (DTM) 5m",
    merge_tiles=True,
    **kwargs,
) -> xr.DataArray:
    """Download ahn data within an extent.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent xmin, xmax, ymin, ymax.
    identifier : str, optional
        The identifier determines the AHN-version, the resolution and the type of height
        data. Possible values are (casing is important):
            AHN1: 'AHN1 maaiveldmodel (DTM) 5m'
            AHN2: 'AHN2 maaiveldmodel (DTM) ½m, geïnterpoleerd',
                  'AHN2 maaiveldmodel (DTM) ½m',
                  'AHN2 DSM ½m',
                  'AHN2 maaiveldmodel (DTM) 5m'
            AHN3: 'AHN3 maaiveldmodel (DTM) ½m',
                  'AHN3 DSM ½m',
                  'AHN3 maaiveldmodel (DTM) 5m',
                  'AHN3 DSM 5m'
            AHN4: 'AHN4 maaiveldmodel (DTM) ½m',
                  'AHN4 DSM ½m',
                  'AHN4 maaiveldmodel (DTM) 5m',
                  'AHN4 DSM 5m'
            AHN5: 'AHN5 maaiveldmodel (DTM) 5m',
                  'AHN5 DSM 5m',
                  'AHN5 maaiveldmodel (DTM) ½m',
                  'AHN5 DSM ½m'
    merge_tiles : bool, optional
        If True, the function returns a merged DataArray. If False, the function
        returns a list of DataArrays with the original tiles. The default is True.

    Returns
    -------
    ahn_da : xr.DataArray
        DataArray with the ahn variable.
    """

    ahn_da = _download_ahn_ellipsis(
        extent, identifier=identifier, merge_tiles=merge_tiles, **kwargs
    )
    if not merge_tiles:
        return ahn_da

    ahn_da.attrs["source"] = identifier
    ahn_da.attrs["date"] = dt.datetime.now().strftime("%Y%m%d")
    ahn_da.attrs["units"] = "mNAP"

    return ahn_da


@cache.cache_netcdf(coords_2d=True)
def discretize_ahn(
    ds: xr.Dataset, ahn_da: xr.DataArray, method: str = "average"
) -> xr.Dataset:
    """Discretize ahn data to model the model grid.

    Parameters
    ----------
    ds : xr.Dataset
        dataset with the model information.
    ahn_da : xr.DataArray
        ahn data within model extent.
    method : str, optional
        Method used to resample ahn to grid of ds. See documentation of
        nlmod.resample.structured_da_to_ds for possible values. The default is
        'average'.

    Returns
    -------
    ds_out : xr.Dataset
        Dataset with the ahn variable.
    """
    if ds is not None:
        ahn_da = structured_da_to_ds(ahn_da, ds, method=method)

    ds_out = get_ds_empty(ds, keep_coords=("y", "x"))
    ds_out["ahn"] = ahn_da

    return ds_out


@cache.cache_netcdf(coords_2d=True)
def get_ahn(
    ds: xr.Dataset | None = None,
    identifier: str = "AHN4 maaiveldmodel (DTM) 5m",
    method: str = "average",
    extent: list[float] | None = None,
    merge_tiles: bool = True,
    **kwargs,
) -> xr.Dataset:
    """Get a model dataset with ahn variable.

    Parameters
    ----------
    ds : xr.Dataset, optional
        dataset with the model information. Using ds=None is deprecated, instead use
        nlmod.read.ahn.download_ahn().
    identifier : str, optional
        The identifier determines the AHN-version, the resolution and the type of height
        data. Possible values are (casing is important):
            AHN1: 'AHN1 maaiveldmodel (DTM) 5m'
            AHN2: 'AHN2 maaiveldmodel (DTM) ½m, geïnterpoleerd',
                  'AHN2 maaiveldmodel (DTM) ½m',
                  'AHN2 DSM ½m',
                  'AHN2 maaiveldmodel (DTM) 5m'
            AHN3: 'AHN3 maaiveldmodel (DTM) ½m',
                  'AHN3 DSM ½m',
                  'AHN3 maaiveldmodel (DTM) 5m',
                  'AHN3 DSM 5m'
            AHN4: 'AHN4 maaiveldmodel (DTM) ½m',
                  'AHN4 DSM ½m',
                  'AHN4 maaiveldmodel (DTM) 5m',
                  'AHN4 DSM 5m'
            AHN5: 'AHN5 maaiveldmodel (DTM) 5m',
                  'AHN5 DSM 5m',
                  'AHN5 maaiveldmodel (DTM) ½m',
                  'AHN5 DSM ½m'
        The default is 'AHN4 maaiveldmodel (DTM) 5m'.
    method : str, optional
        Method used to resample ahn to grid of ds. See documentation of
        nlmod.resample.structured_da_to_ds for possible values. The default is
        'average'.
    extent : list, tuple or np.array, optional
        extent xmin, xmax, ymin, ymax. Only used if ds is None. The default is None.

    Returns
    -------
    ds_out : xr.Dataset
        Dataset with the ahn variable.
    """
    if ds is None:
        warnings.warn(
            "calling 'get_ahn' with ds=None is deprecated and will raise an error in the "
            "future. Use 'nlmod.read.ahn.download_ahn' to get the ahn within an extent",
            DeprecationWarning,
        )

    if extent is None and ds is not None:
        extent = get_extent(ds)

    ahn_da = download_ahn(
        extent=extent, identifier=identifier, merge_tiles=merge_tiles, **kwargs
    )
    # this is probably redundant when we have the 'download_ahn' function
    if not merge_tiles:
        return ahn_da

    if ds is None:
        return ahn_da
    else:
        return discretize_ahn(ds, ahn_da, method=method)


def get_ahn_at_point(
    x: float,
    y: float,
    buffer: float = 0.75,
    return_da: bool = False,
    return_mean: bool = False,
    identifier: str = "dsm_05m",
    res: float = 0.5,
    **kwargs,
) -> float:
    """Get the height of the surface level at a certain point, defined by x and y.

    .. deprecated:: 0.10.0
        `get_ahn_at_point` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn_at_point` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    x : float
        The x-coordinate fo the point.
    y : float
        The y-coordinate fo the point..
    buffer : float, optional
        The buffer around x and y that is downloaded. The default is 0.75.
    return_da : bool, optional
        Return the downloaded DataArray when True. The default is False.
    return_mean : bool, optional
        Resturn the mean of all non-nan pixels within buffer. Return the center pixel
        when False. The default is False.
    identifier : str, optional
        The identifier passed onto download_latest_ahn_from_wcs. The default is "dsm_05m".
    res : float, optional
        The resolution that is passed onto download_latest_ahn_from_wcs. The default is 0.5.
    **kwargs : dict
        kwargs are passed onto the method download_latest_ahn_from_wcs.

    Returns
    -------
    float
        The surface level value at the requested point.
    """

    warnings.warn(
        "'get_ahn_at_point' is deprecated and will eventually be removed, "
        "please use nlmod.read.ahn.download_ahn_at_point() in the future.",
        DeprecationWarning,
    )

    return download_ahn_at_point(
        x, y, buffer, return_da, return_mean, identifier, res, **kwargs
    )


def download_ahn_at_point(
    x: float,
    y: float,
    buffer: float = 0.75,
    return_da: bool = False,
    return_mean: bool = False,
    identifier: str = "dsm_05m",
    res: float = 0.5,
    **kwargs,
) -> float:
    """Get the height of the surface level at a certain point, defined by x and y.

    Parameters
    ----------
    x : float
        The x-coordinate fo the point.
    y : float
        The y-coordinate fo the point..
    buffer : float, optional
        The buffer around x and y that is downloaded. The default is 0.75.
    return_da : bool, optional
        Return the downloaded DataArray when True. The default is False.
    return_mean : bool, optional
        Resturn the mean of all non-nan pixels within buffer. Return the center pixel
        when False. The default is False.
    identifier : str, optional
        The identifier passed onto download_latest_ahn_from_wcs. The default is "dsm_05m".
    res : float, optional
        The resolution that is passed onto download_latest_ahn_from_wcs. The default is 0.5.
    **kwargs : dict
        kwargs are passed onto the method download_latest_ahn_from_wcs.

    Returns
    -------
    float
        The surface level value at the requested point.
    """
    extent = [x - buffer, x + buffer, y - buffer, y + buffer]
    ahn = download_latest_ahn_from_wcs(extent, identifier=identifier, res=res, **kwargs)
    if return_da:
        # return a DataArray
        return ahn
    if return_mean:
        # return the mean (usefull when there are NaN's near the center)
        return float(ahn.mean())
    else:
        # return the center pixel
        return ahn.data[int((ahn.shape[0] - 1) / 2), int((ahn.shape[1] - 1) / 2)]


def get_ahn_along_line(
    line: shapely.LineString,
    ahn: xr.DataArray | None = None,
    dx: float | None = None,
    num: int | None = None,
    method: str = "linear",
    plot: bool = False,
) -> xr.DataArray:
    """Get the height of the surface level along a line.

    .. deprecated:: 0.10.0
        `get_ahn_along_line` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn_along_line` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    line : shapely.LineString
        The line along which the surface level is calculated.
    ahn : xr.DataArray, optional
        The 2d DataArray containing surface level values. If None, ahn4-values are
        downloaded from the web. The default is None.
    dx : float, optional
        The distance between the points along the line at which the surface level is
        calculated. Only used when num is None. When dx is None, it is set to the
        resolution of ahn. The default is None.
    num : int, optional
        If not None, the surface level is calculated at num equally spaced points along
        the line. The default is None.
    method : string, optional
        The method to interpolate the 2d surface level values to the points along the
        line. The default is "linear".
    plot : bool, optional
        if True, plot the 2d surface level, the line and the calculated heights. The
        default is False.

    Returns
    -------
    z : xr.DataArray
        A DataArray with dimension s, containing surface level values along the line.
    """
    warnings.warn(
        "'get_ahn_along_line' is deprecated and will eventually be removed, "
        "please use nlmod.read.ahn.download_ahn_along_line() in the future.",
        DeprecationWarning,
    )

    return download_ahn_along_line(line, ahn, dx, num, method, plot)


def download_ahn_along_line(
    line: shapely.LineString,
    ahn: xr.DataArray | None = None,
    dx: float | None = None,
    num: int | None = None,
    method: str = "linear",
    plot: bool = False,
) -> xr.DataArray:
    """Get the height of the surface level along a line.

    Parameters
    ----------
    line : shapely.LineString
        The line along which the surface level is calculated.
    ahn : xr.DataArray, optional
        The 2d DataArray containing surface level values. If None, ahn4-values are
        downloaded from the web. The default is None.
    dx : float, optional
        The distance between the points along the line at which the surface level is
        calculated. Only used when num is None. When dx is None, it is set to the
        resolution of ahn. The default is None.
    num : int, optional
        If not None, the surface level is calculated at num equally spaced points along
        the line. The default is None.
    method : string, optional
        The method to interpolate the 2d surface level values to the points along the
        line. The default is "linear".
    plot : bool, optional
        if True, plot the 2d surface level, the line and the calculated heights. The
        default is False.

    Returns
    -------
    z : xr.DataArray
        A DataArray with dimension s, containing surface level values along the line.
    """
    if ahn is None:
        bbox = line.bounds
        extent = [bbox[0], bbox[2], bbox[1], bbox[3]]
        ahn = get_ahn4(extent)
    if num is not None:
        s = np.linspace(0.0, line.length, num)
    else:
        if dx is None:
            dx = float(ahn.x[1] - ahn.x[0])
        s = np.arange(0.0, line.length, dx)

    x, y = zip(*[p.xy for p in line.interpolate(s)])

    x = np.array(x)[:, 0]
    y = np.array(y)[:, 0]

    x = xr.DataArray(x, dims="s", coords={"s": s})
    y = xr.DataArray(y, dims="s", coords={"s": s})
    z = ahn.interp(x=x, y=y, method=method)

    if plot:
        _, ax = plt.subplots(figsize=(10, 10))
        ahn.plot(ax=ax)
        gpd.GeoDataFrame(geometry=[line]).plot(ax=ax)

        _, ax = plt.subplots(figsize=(10, 10))
        z.plot(ax=ax)
    return z


@cache.cache_netcdf()
def get_latest_ahn_from_wcs(
    extent: list[float] = None,
    identifier: Literal["dsm_05m", "dtm_05m"] = "dsm_05m",
    res: float | None = None,
    version: str = "1.0.0",
    fmt: str = "image/tiff",
    crs: str = "EPSG:28992",
    maxsize: int = 2000,
) -> xr.DataArray:
    """Get the latest AHN from the wcs service.

    .. deprecated:: 0.10.0
        `get_latest_ahn_from_wcs` will be removed in nlmod 1.0.0, it is replaced by
        `download_latest_ahn_from_wcs` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array, optional
        extent. The default is None.
    identifier : str, optional
        Possible values for identifier are:
            'dsm_05m'
            'dtm_05m'
        The default is 'dsm_05m'.
        the identifier contains resolution and type info:
        - 'dtm' is only surface level (maaiveld), 'dsm' has other surfaces
        such as buildings.
        - 5m or 05m is a resolution of 5x5 or 0.5x0.5 meter.
    res : float, optional
        resolution of ahn raster. If None the resolution is inferred from the
        identifier. The default is None.
    version : str, optional
        version of wcs service, options are '1.0.0' and '2.0.1'.
        The default is '1.0.0'.
    fmt : str, optional
        geotif format . The default is 'GEOTIFF_FLOAT32'.
    crs : str, optional
        coördinate reference system. The default is 'EPSG:28992'.
    maxsize : float, optional
        maximum number of cells in x or y direction. The default is
        2000.

    Returns
    -------
    xr.DataArray
    """
    warnings.warn(
        "'get_latest_ahn_from_wcs' is deprecated and will eventually be removed, "
        "please use 'nlmod.read.ahn.download_latest_ahn_from_wcs()' in the future.",
        DeprecationWarning,
    )

    return download_latest_ahn_from_wcs(
        extent, identifier, res, version, fmt, crs, maxsize
    )


@cache.cache_netcdf()
def download_latest_ahn_from_wcs(
    extent: list[float] = None,
    identifier: Literal["dsm_05m", "dtm_05m"] = "dsm_05m",
    res: float | None = None,
    version: str = "1.0.0",
    fmt: str = "image/tiff",
    crs: str = "EPSG:28992",
    maxsize: int = 2000,
) -> xr.DataArray:
    """Get the latest AHN from the wcs service.

    Parameters
    ----------
    extent : list, tuple or np.array, optional
        extent. The default is None.
    identifier : str, optional
        Possible values for identifier are:
            'dsm_05m'
            'dtm_05m'
        The default is 'dsm_05m'.
        the identifier contains resolution and type info:
        - 'dtm' is only surface level (maaiveld), 'dsm' has other surfaces
        such as buildings.
        - 5m or 05m is a resolution of 5x5 or 0.5x0.5 meter.
    res : float, optional
        resolution of ahn raster. If None the resolution is inferred from the
        identifier. The default is None.
    version : str, optional
        version of wcs service, options are '1.0.0' and '2.0.1'.
        The default is '1.0.0'.
    fmt : str, optional
        geotif format . The default is 'GEOTIFF_FLOAT32'.
    crs : str, optional
        coördinate reference system. The default is 'EPSG:28992'.
    maxsize : float, optional
        maximum number of cells in x or y direction. The default is
        2000.

    Returns
    -------
    xr.DataArray
    """
    url = "https://service.pdok.nl/rws/ahn/wcs/v1_0?SERVICE=WCS&request=GetCapabilities"

    if isinstance(extent, xr.DataArray):
        extent = tuple(extent.values)

    # check resolution
    if res is None:
        if "05m" in identifier.split("_")[1]:
            res = 0.5
        elif "5m" in identifier.split("_")[1]:
            logger.warning(
                "5 meter resolution is no langer available via wcs, try "
                "nlmod.read.get_ahn4 to obtain ahn with a 5m resolution. For "
                "more info see: "
                "https://www.pdok.nl/-/nieuwe-versie-ahn-beschikbaar-via-pdok"
            )
            raise ValueError(
                "5 meter resolution no longer available via wcs use nlmod.read.get_ahn4"
            )
        else:
            raise ValueError("could not infer resolution from identifier")

    da = wcs(
        url,
        extent,
        res,
        identifier=identifier,
        version=version,
        fmt=fmt,
        crs=crs,
        maxsize=maxsize,
    )
    return da


def get_ahn2_tiles(extent: list[float] | None = None) -> gpd.GeoDataFrame:
    """Get the tiles (kaartbladen) of AHN3 as a GeoDataFrame.

    The links in the tiles are cuurently incorrect. Thereore get_ahn3_tiles is used in
    get_ahn2 and get_ahn1, as the tiles from get_ahn3_tiles also contain information
    about the tiles of ahn1 and ahn2
    """
    url = "https://services.arcgis.com/nSZVuSZjHpEZZbRo/arcgis/rest/services/Kaartbladen_AHN2/FeatureServer"
    layer = 0
    gdf = arcrest(url, layer, extent)
    if not gdf.empty:
        gdf = gdf.set_index("Kaartblad")
    return gdf


def get_ahn3_tiles(extent: list[float] | None = None) -> gpd.GeoDataFrame:
    """Get the tiles (kaartbladen) of AHN3 as a GeoDataFrame."""
    url = "https://services.arcgis.com/nSZVuSZjHpEZZbRo/arcgis/rest/services/Kaartbladen_AHN3/FeatureServer"
    layer = 0
    gdf = arcrest(url, layer, extent)
    if not gdf.empty:
        gdf = gdf.set_index("Kaartblad")
    return gdf


def get_ahn4_tiles(extent: list[float] | None = None) -> gpd.GeoDataFrame:
    """Get the tiles (kaartbladen) of AHN4 as a GeoDataFrame with download links."""
    url = "https://services.arcgis.com/nSZVuSZjHpEZZbRo/arcgis/rest/services/Kaartbladen_AHN4/FeatureServer"
    layer = 0
    gdf = arcrest(url, layer, extent)
    if not gdf.empty:
        gdf = gdf.set_index("Name")
    return gdf


def _download_tiles_ellipsis(
    extent: list[float] | None = None,
    crs: int = 28992,
    timeout: float = 120.0,
    base_url: str = "https://api.ellipsis-drive.com/v3",
    path_id: str = "a9d410ad-a2f6-404c-948a-fdf6b43e77a6",
    timestamp_id: str = "05931403-2510-43af-9cc3-f60a066d4482",
) -> gpd.GeoDataFrame:
    url = f"{base_url}/path/{path_id}/vector/timestamp/{timestamp_id}/listFeatures"

    r = requests.get(url, timeout=timeout)
    if not r.ok:
        raise (HTTPError(f"Request not successful: {r.url}"))
    gdf = gpd.GeoDataFrame.from_features(r.json()["result"]["features"], crs=4326)
    gdf = gdf.to_crs(crs)
    # remove small digits becuase of crs-transformation
    gdf.geometry = gdf.geometry.apply(_round_coordinates, ndigits=0)

    gdf = gdf.set_index("AHN")

    if extent is not None:
        gdf = gdf.loc[gdf.intersection(extent_to_polygon(extent)).area > 0]
    return gdf


def _get_tiles_from_file(
    fname: str,
    extent: list[float] | None = None,
    crs: int = 28992,
) -> gpd.GeoDataFrame:
    if crs != 28992:
        raise ValueError("Only crs 28992 is supported")

    gdf = gpd.read_file(fname)

    # remove small digits becuase of crs-transformation
    gdf = gdf.set_index("AHN")

    if extent is not None:
        gdf = gdf.loc[gdf.intersection(extent_to_polygon(extent)).area > 0]

    return gdf


def _round_coordinates(geom: shapely.Geometry, ndigits: int = 2) -> shapely.Geometry:
    def _round_coords(x, y, z=None):
        x = round(x, ndigits)
        y = round(y, ndigits)

        if z is not None:
            z = round(x, ndigits)
            return (x, y, z)
        else:
            return (x, y)

    return shapely.ops.transform(_round_coords, geom)


@cache.cache_netcdf()
def get_ahn1(
    extent: list[float],
    identifier: Literal["AHN1 maaiveldmodel (DTM) 5m"] = "AHN1 maaiveldmodel (DTM) 5m",
    as_data_array: bool | None = None,
    **kwargs,
) -> xr.DataArray:
    """Download AHN1.

    .. deprecated:: 0.10.0
          `get_ahn1` will be removed in nlmod 1.0.0, it is replaced by
          `download_ahn1` because of new naming convention
          https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. The only
        allowed value is 'AHN1 maaiveldmodel (DTM) 5m'. The default is
        "AHN1 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    warnings.warn(
        "'get_ahn1' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn1' instead",
        DeprecationWarning,
    )
    return download_ahn1(extent, identifier, as_data_array, **kwargs)


@cache.cache_netcdf()
def download_ahn1(
    extent: list[float],
    identifier: Literal["AHN1 maaiveldmodel (DTM) 5m"] = "AHN1 maaiveldmodel (DTM) 5m",
    as_data_array: bool | None = None,
    **kwargs,
) -> xr.DataArray:
    """Download AHN1.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. The only
        allowed value is 'AHN1 maaiveldmodel (DTM) 5m'. The default is
        "AHN1 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    _assert_as_data_array_is_none(as_data_array)
    da = _download_ahn_ellipsis(extent, identifier, **kwargs)
    if "merge_tiles" in kwargs and kwargs["merge_tiles"]:
        return da
    # original data is in cm. Convert the data to m, which is the unit of other ahns
    da = da / 100
    return da


@cache.cache_netcdf()
def get_ahn1_legacy(
    extent: list[float],
    identifier: Literal["ahn1_5m"] = "ahn1_5m",
    as_data_array: bool = True,
) -> xr.DataArray:
    """Download AHN1.

    .. deprecated:: 0.10.0
        `get_ahn1_legacy` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn1_legacy` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. The only
        allowed value is 'ahn1_5m'. The default is "ahn1_5m".
    as_data_array : bool, optional
        return the data as as xarray DataArray if true. The default is True.

    Returns
    -------
    xr.DataArray or MemoryFile
        DataArray (if as_data_array is True) or Rasterio MemoryFile of the AHN
    """
    warnings.warn(
        "'get_ahn1_legacy' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn1_legacy' instead",
        DeprecationWarning,
    )
    return download_ahn1_legacy(extent, identifier, as_data_array)


@cache.cache_netcdf()
def download_ahn1_legacy(
    extent: list[float],
    identifier: Literal["ahn1_5m"] = "ahn1_5m",
    as_data_array: bool = True,
) -> xr.DataArray:
    """Download AHN1.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. The only
        allowed value is 'ahn1_5m'. The default is "ahn1_5m".
    as_data_array : bool, optional
        return the data as as xarray DataArray if true. The default is True.

    Returns
    -------
    xr.DataArray or MemoryFile
        DataArray (if as_data_array is True) or Rasterio MemoryFile of the AHN
    """
    # tiles are equal to that of ahn3
    tiles = get_ahn3_tiles(extent)
    da = _download_and_combine_tiles(tiles, identifier, extent, as_data_array)
    if as_data_array:
        # original data is in cm. Convert the data to m, which is the unit of other ahns
        da = da / 100
    return da


@cache.cache_netcdf()
def get_ahn2(
    extent: list[float],
    identifier: Literal[
        "AHN2 maaiveldmodel (DTM) ½m, geïnterpoleerd",
        "AHN2 maaiveldmodel (DTM) ½m",
        "AHN2 DSM ½m",
        "AHN2 maaiveldmodel (DTM) 5m",
    ] = "AHN2 maaiveldmodel (DTM) 5m",
    as_data_array: bool | None = None,
    **kwargs,
) -> xr.DataArray:
    """Download AHN2.

    .. deprecated:: 0.10.0
        `get_ahn2` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn2` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN2 maaiveldmodel (DTM) ½m, geïnterpoleerd',
        'AHN2 maaiveldmodel (DTM) ½m', 'AHN2 DSM ½m', and
        'AHN2 maaiveldmodel (DTM) 5m'. The default is
        "AHN2 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    warnings.warn(
        "'get_ahn2' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn2' instead",
        DeprecationWarning,
    )
    return download_ahn2(extent, identifier, as_data_array, **kwargs)


@cache.cache_netcdf()
def download_ahn2(
    extent: list[float],
    identifier: Literal[
        "AHN2 maaiveldmodel (DTM) ½m, geïnterpoleerd",
        "AHN2 maaiveldmodel (DTM) ½m",
        "AHN2 DSM ½m",
        "AHN2 maaiveldmodel (DTM) 5m",
    ] = "AHN2 maaiveldmodel (DTM) 5m",
    as_data_array: bool | None = None,
    **kwargs,
) -> xr.DataArray:
    """Download AHN2.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN2 maaiveldmodel (DTM) ½m, geïnterpoleerd',
        'AHN2 maaiveldmodel (DTM) ½m', 'AHN2 DSM ½m', and
        'AHN2 maaiveldmodel (DTM) 5m'. The default is
        "AHN2 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    _assert_as_data_array_is_none(as_data_array)
    return _download_ahn_ellipsis(extent, identifier, **kwargs)


@cache.cache_netcdf()
def get_ahn2_legacy(
    extent: list[float],
    identifier: Literal[
        "ahn2_05m_i", "ahn2_05m_n", "ahn2_05m_r", "ahn2_5m"
    ] = "ahn2_5m",
    as_data_array: bool = True,
) -> xr.DataArray | MemoryFile:
    """Download AHN2.

    .. deprecated:: 0.10.0
        `get_ahn2_legacy` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn2_legacy` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'ahn2_05m_i', 'ahn2_05m_n', 'ahn2_05m_r' and 'ahn2_5m'. The default
        is "ahn2_5m".
    as_data_array : bool, optional
        return the data as as xarray DataArray if true. The default is True.

    Returns
    -------
    xr.DataArray or MemoryFile
        DataArray (if as_data_array is True) or Rasterio MemoryFile of the AHN
    """
    warnings.warn(
        "'get_ahn2_legacy' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn2_legacy' instead",
        DeprecationWarning,
    )
    return download_ahn2_legacy(extent, identifier, as_data_array)


@cache.cache_netcdf()
def download_ahn2_legacy(
    extent: list[float],
    identifier: Literal[
        "ahn2_05m_i", "ahn2_05m_n", "ahn2_05m_r", "ahn2_5m"
    ] = "ahn2_5m",
    as_data_array: bool = True,
) -> xr.DataArray | MemoryFile:
    """Download AHN2.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'ahn2_05m_i', 'ahn2_05m_n', 'ahn2_05m_r' and 'ahn2_5m'. The default
        is "ahn2_5m".
    as_data_array : bool, optional
        return the data as as xarray DataArray if true. The default is True.

    Returns
    -------
    xr.DataArray or MemoryFile
        DataArray (if as_data_array is True) or Rasterio MemoryFile of the AHN
    """
    # tiles are equal to that of ahn3
    tiles = get_ahn3_tiles(extent)
    return _download_and_combine_tiles(tiles, identifier, extent, as_data_array)


@cache.cache_netcdf()
def get_ahn3(
    extent: list[float],
    identifier: Literal[
        "AHN3 maaiveldmodel (DTM) ½m",
        "AHN3 DSM ½m",
        "AHN3 maaiveldmodel (DTM) 5m",
        "AHN3 DSM 5m",
    ] = "AHN3 maaiveldmodel (DTM) 5m",
    as_data_array: bool | None = None,
    **kwargs,
) -> xr.DataArray:
    """Download AHN3.

    .. deprecated:: 0.10.0
        `get_ahn3` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn3` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN3 maaiveldmodel (DTM) ½m', 'AHN3 DSM ½m',
        'AHN3 maaiveldmodel (DTM) 5m', and 'AHN3 DSM 5m'. The default is
        "AHN3 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    warnings.warn(
        "'get_ahn3' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn3' instead",
        DeprecationWarning,
    )
    return download_ahn3(extent, identifier, as_data_array, **kwargs)


@cache.cache_netcdf()
def download_ahn3(
    extent: list[float],
    identifier: Literal[
        "AHN3 maaiveldmodel (DTM) ½m",
        "AHN3 DSM ½m",
        "AHN3 maaiveldmodel (DTM) 5m",
        "AHN3 DSM 5m",
    ] = "AHN3 maaiveldmodel (DTM) 5m",
    as_data_array: bool | None = None,
    **kwargs,
) -> xr.DataArray:
    """Download AHN3.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN3 maaiveldmodel (DTM) ½m', 'AHN3 DSM ½m',
        'AHN3 maaiveldmodel (DTM) 5m', and 'AHN3 DSM 5m'. The default is
        "AHN3 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    _assert_as_data_array_is_none(as_data_array)
    return _download_ahn_ellipsis(extent, identifier, **kwargs)


@cache.cache_netcdf()
def get_ahn3_legacy(
    extent: list[float],
    identifier: Literal[
        "AHN3_05m_DSM", "AHN3_05m_DTM", "AHN3_5m_DSM", "AHN3_5m_DTM"
    ] = "AHN3_5m_DTM",
    as_data_array: bool = True,
) -> xr.DataArray | MemoryFile:
    """Download AHN3.

    .. deprecated:: 0.10.0
        `get_ahn3_legacy` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn3_legacy` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN3_05m_DSM', 'AHN3_05m_DTM', 'AHN3_5m_DSM' and 'AHN3_5m_DTM'. The
        default is "AHN3_5m_DTM".
    as_data_array : bool, optional
        return the data as as xarray DataArray if true. The default is True.

    Returns
    -------
    xr.DataArray or MemoryFile
        DataArray (if as_data_array is True) or Rasterio MemoryFile of the AHN
    """
    warnings.warn(
        "'get_ahn3_legacy' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn3_legacy' instead",
        DeprecationWarning,
    )
    return download_ahn3_legacy(extent, identifier, as_data_array)


@cache.cache_netcdf()
def download_ahn3_legacy(
    extent: list[float],
    identifier: Literal[
        "AHN3_05m_DSM", "AHN3_05m_DTM", "AHN3_5m_DSM", "AHN3_5m_DTM"
    ] = "AHN3_5m_DTM",
    as_data_array: bool = True,
) -> xr.DataArray | MemoryFile:
    """Download AHN3.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN3_05m_DSM', 'AHN3_05m_DTM', 'AHN3_5m_DSM' and 'AHN3_5m_DTM'. The
        default is "AHN3_5m_DTM".
    as_data_array : bool, optional
        return the data as as xarray DataArray if true. The default is True.

    Returns
    -------
    xr.DataArray or MemoryFile
        DataArray (if as_data_array is True) or Rasterio MemoryFile of the AHN
    """
    tiles = get_ahn3_tiles(extent)
    return _download_and_combine_tiles(tiles, identifier, extent, as_data_array)


@cache.cache_netcdf()
def get_ahn4(
    extent: list[float],
    identifier: Literal[
        "AHN4 maaiveldmodel (DTM) ½m",
        "AHN4 DSM ½m",
        "AHN4 maaiveldmodel (DTM) 5m",
        "AHN4 DSM 5m",
    ] = "AHN4 maaiveldmodel (DTM) 5m",
    as_data_array: bool | None = None,
    **kwargs,
) -> xr.DataArray:
    """Download AHN4.

    .. deprecated:: 0.10.0
        `get_ahn4` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn4` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN4 maaiveldmodel (DTM) ½m', 'AHN4 DSM ½m',
        'AHN4 maaiveldmodel (DTM) 5m', and 'AHN4 DSM 5m'. The default is
        "AHN4 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    warnings.warn(
        "'get_ahn4' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn4' instead",
        DeprecationWarning,
    )
    return download_ahn4(extent, identifier, as_data_array, **kwargs)


@cache.cache_netcdf()
def download_ahn4(
    extent: list[float],
    identifier: Literal[
        "AHN4 maaiveldmodel (DTM) ½m",
        "AHN4 DSM ½m",
        "AHN4 maaiveldmodel (DTM) 5m",
        "AHN4 DSM 5m",
    ] = "AHN4 maaiveldmodel (DTM) 5m",
    as_data_array: bool | None = None,
    **kwargs,
) -> xr.DataArray:
    """Download AHN4.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN4 maaiveldmodel (DTM) ½m', 'AHN4 DSM ½m',
        'AHN4 maaiveldmodel (DTM) 5m', and 'AHN4 DSM 5m'. The default is
        "AHN4 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    _assert_as_data_array_is_none(as_data_array)
    return _download_ahn_ellipsis(extent, identifier, **kwargs)


@cache.cache_netcdf()
def get_ahn4_legacy(
    extent: list[float],
    identifier: Literal[
        "AHN4_DTM_05m", "AHN4_DSM_05m", "AHN4_DTM_5m", "AHN4_DSM_5m"
    ] = "AHN4_DTM_5m",
    as_data_array: bool = True,
) -> xr.DataArray | MemoryFile:
    """Download AHN4.

    .. deprecated:: 0.10.0
        `get_ahn4_legacy` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn4_legacy` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN4_DTM_05m', 'AHN4_DTM_5m', 'AHN4_DSM_05m' and 'AHN4_DSM_5m'. The
        default is "AHN4_DTM_5m".
    as_data_array : bool, optional
        return the data as as xarray DataArray if true. The default is True.

    Returns
    -------
    xr.DataArray or MemoryFile
        DataArray (if as_data_array is True) or Rasterio MemoryFile of the AHN
    """
    warnings.warn(
        "'get_ahn4_legacy' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn4_legacy' instead",
        DeprecationWarning,
    )
    return download_ahn4_legacy(extent, identifier, as_data_array)


@cache.cache_netcdf()
def download_ahn4_legacy(
    extent: list[float],
    identifier: Literal[
        "AHN4_DTM_05m", "AHN4_DSM_05m", "AHN4_DTM_5m", "AHN4_DSM_5m"
    ] = "AHN4_DTM_5m",
    as_data_array: bool = True,
) -> xr.DataArray | MemoryFile:
    """Download AHN4.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN4_DTM_05m', 'AHN4_DTM_5m', 'AHN4_DSM_05m' and 'AHN4_DSM_5m'. The
        default is "AHN4_DTM_5m".
    as_data_array : bool, optional
        return the data as as xarray DataArray if true. The default is True.

    Returns
    -------
    xr.DataArray or MemoryFile
        DataArray (if as_data_array is True) or Rasterio MemoryFile of the AHN
    """
    tiles = get_ahn4_tiles(extent)
    return _download_and_combine_tiles(tiles, identifier, extent, as_data_array)


@cache.cache_netcdf()
def get_ahn5(
    extent: list[float],
    identifier: Literal[
        "AHN5 maaiveldmodel (DTM) 5m",
        "AHN5 DSM 5m",
        "AHN5 maaiveldmodel (DTM) ½m",
        "AHN5 DSM ½m",
    ] = "AHN5 maaiveldmodel (DTM) 5m",
    **kwargs,
) -> xr.DataArray:
    """Download AHN5.

    .. deprecated:: 0.10.0
        `get_ahn5` will be removed in nlmod 1.0.0, it is replaced by
        `download_ahn5` because of new naming convention
        https://github.com/gwmod/nlmod/issues/47

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN5 maaiveldmodel (DTM) 5m', 'AHN5 DSM 5m',
        'AHN5 maaiveldmodel (DTM) ½m', and 'AHN5 DSM ½m'. The default is
        "AHN5 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    warnings.warn(
        "'get_ahn5' is deprecated and will be removed in a future version. "
        "Use 'nlmod.read.ahn.download_ahn5' instead",
        DeprecationWarning,
    )
    return download_ahn5(extent, identifier, **kwargs)


@cache.cache_netcdf()
def download_ahn5(
    extent: list[float],
    identifier: Literal[
        "AHN5 maaiveldmodel (DTM) 5m",
        "AHN5 DSM 5m",
        "AHN5 maaiveldmodel (DTM) ½m",
        "AHN5 DSM ½m",
    ] = "AHN5 maaiveldmodel (DTM) 5m",
    **kwargs,
) -> xr.DataArray:
    """Download AHN5.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str, optional
        The identifier determines the resolution and the type of height data. Possible
        values are 'AHN5 maaiveldmodel (DTM) 5m', 'AHN5 DSM 5m',
        'AHN5 maaiveldmodel (DTM) ½m', and 'AHN5 DSM ½m'. The default is
        "AHN5 maaiveldmodel (DTM) 5m".

    Returns
    -------
    xr.DataArray
        DataArray of the AHN
    """
    return _download_ahn_ellipsis(extent, identifier, **kwargs)


def _update_ellipsis_tiles_in_data() -> None:
    tiles = _download_tiles_ellipsis()
    pathname = os.path.join(NLMOD_DATADIR, "ahn")
    if not os.path.isdir(pathname):
        os.makedirs(pathname)
    fname = os.path.join(pathname, "ellipsis_tiles.geojson")
    tiles.to_file(fname)


@cache.cache_netcdf()
def _download_ahn_ellipsis(
    extent: list[float], identifier: str, merge_tiles: bool = True, **kwargs
) -> xr.DataArray | list[xr.DataArray]:
    """Download and merge AHN tiles from the ellipsis.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    identifier : str
        The identifier determines the AHN-version, the resolution and the type of height
        data.
    merge_tiles : bool, optional
        If True, the function returns a merged DataArray. If False, the function
        returns a list of DataArrays with the original tiles. The default is True.

    Returns
    -------
    xr.DataArray
        DataArray of the AHN

    """
    fname = os.path.join(NLMOD_DATADIR, "ahn", "ellipsis_tiles.geojson")
    tiles = _get_tiles_from_file(fname, extent=extent, **kwargs)

    identifier = _rename_identifier(identifier)
    if identifier not in tiles.columns:
        raise (ValueError(f"Unknown ahn-identifier: {identifier}"))
    tiles = tiles[~tiles[identifier].isna()]
    das = []
    for tile in tqdm(tiles.index, desc=f"Downloading tiles of {identifier}"):
        url = tiles.at[tile, identifier]
        if url == "None":
            continue
        if url.endswith(".zip"):
            path = url.split("/")[-1].replace(".zip", ".TIF")
            if path.lower().endswith(".tif.tif"):
                path = path[:-4]
            da = rioxarray.open_rasterio(f"zip+{url}!/{path}", mask_and_scale=True)
        else:
            da = rioxarray.open_rasterio(url, mask_and_scale=True)
        da = da.sel(x=slice(extent[0], extent[1]), y=slice(extent[3], extent[2]))
        das.append(da)

    if len(das) == 0:
        raise (ValueError("No data found within extent"))

    if merge_tiles:
        da = merge_arrays(das)
        if da.dims[0] == "band":
            da = da[0].drop_vars("band")
        if "_FillValue" in da.attrs:
            del da.attrs["_FillValue"]
        return da
    return das


def _download_and_combine_tiles(
    tiles: gpd.GeoDataFrame, identifier: str, extent: list[float], as_data_array: bool
) -> xr.DataArray | MemoryFile:
    """Internal method to download and combine ahn-data."""
    if tiles.empty:
        raise (Exception(f"{identifier} has no data for requested extent"))
    datasets = []
    for name in tqdm(tiles.index, desc=f"Downloading tiles of {identifier}"):
        url = tiles.at[name, identifier]
        if isinstance(url, pd.Series):
            logger.warning(
                f"Multiple tiles with the same name: {name}. Choosing the first one."
            )
            url = url.iloc[0]
        path = url.split("/")[-1].replace(".zip", ".TIF")
        if path.lower().endswith(".tif.tif"):
            path = path[:-4]
        datasets.append(rasterio.open(f"zip+{url}!/{path}"))
    memfile = MemoryFile()
    merge.merge(datasets, dst_path=memfile)
    if as_data_array:
        da = rioxarray.open_rasterio(memfile.open(), mask_and_scale=True)[0]
        da = da.sel(x=slice(extent[0], extent[1]), y=slice(extent[3], extent[2]))
        return da
    return memfile


def _rename_identifier(identifier: str) -> str:
    rename = {
        "ahn1_5m": "AHN1 maaiveldmodel (DTM) 5m",
        "AHN1_5M": "AHN1 maaiveldmodel (DTM) 5m",
        "ahn2_05m_i": "AHN2 maaiveldmodel (DTM) ½m, geïnterpoleerd",
        "AHN2_05M_I": "AHN2 maaiveldmodel (DTM) ½m, geïnterpoleerd",
        "ahn2_05m_n": "AHN2 maaiveldmodel (DTM) ½m",
        "AHN2_05M_N": "AHN2 maaiveldmodel (DTM) ½m",
        "ahn2_05m_r": "AHN2 DSM ½m",
        "AHN2_05M_R": "AHN2 DSM ½m",
        "ahn2_5m": "AHN2 maaiveldmodel (DTM) 5m",
        "AHN2_5M_M": "AHN2 maaiveldmodel (DTM) 5m",
        "AHN3_05m_DTM": "AHN3 maaiveldmodel (DTM) ½m",
        "AHN3_05M_M": "AHN3 maaiveldmodel (DTM) ½m",
        "AHN3_05m_DSM": "AHN3 DSM ½m",
        "AHN3_05M_R": "AHN3 DSM ½m",
        "AHN3_5m_DTM": "AHN3 maaiveldmodel (DTM) 5m",
        "AHN3_5M_M": "AHN3 maaiveldmodel (DTM) 5m",
        "AHN3_5m_DSM": "AHN3 DSM 5m",
        "AHN3_5M_R": "AHN3 DSM 5m",
        "AHN4_DTM_05m": "AHN4 maaiveldmodel (DTM) ½m",
        "AHN4_05M_M": "AHN4 maaiveldmodel (DTM) ½m",
        "AHN4_DSM_05m": "AHN4 DSM ½m",
        "AHN4_05M_R": "AHN4 DSM ½m",
        "AHN4_DTM_5m": "AHN4 maaiveldmodel (DTM) 5m",
        "AHN4_5M_M": "AHN4 maaiveldmodel (DTM) 5m",
        "AHN4_DSM_5m": "AHN4 DSM 5m",
        "AHN4_5M_R": "AHN4 DSM 5m",
        "AHN5_5M_M": "AHN5 maaiveldmodel (DTM) 5m",
        "AHN5_5M_R": "AHN5 DSM 5m",
        "AHN5_05M_M": "AHN5 maaiveldmodel (DTM) ½m",
        "AHN5_05M_R": "AHN5 DSM ½m",
    }
    if identifier in rename:
        id_new = rename[identifier]
        logger.warning(f"The identifier {identifier} is deprecated. Rename to {id_new}")
        identifier = id_new
    return identifier


def _assert_as_data_array_is_none(as_data_array: bool | None) -> None:
    if as_data_array is not None:
        raise (
            DeprecationWarning(
                "The as_data_array-argument has been removed from the ahn-"
                "methods, and these methods now allways return a DataArray. "
                "Remove the as_data_array-argument or use the legcay ahn-"
                "methods."
            )
        )
